from collections import namedtuple
import logging
import os
import re

from configobj import ConfigObj

from schema import Schema, And, Use, Optional, Regex

log = logging.getLogger(__name__)

VALID_NAME_RE = re.compile(r'^\S+$')
VALID_NET_ID_RE = re.compile(r'^0x[0-9a-fA-F]{4}$')


def _env_override():
    """Try to find SNAP_NETWORK_PATH env var"""

    try:
        return os.environ['SNAP_NETWORK_PATH']
    except KeyError:
        return None


def user_network_path():
    """Check for the environment var override or use the default"""
    return _env_override() or os.path.expanduser(os.path.join('~', '.snap', 'networks.cfg'))


SCNetwork = namedtuple('SCNetwork', ['channel', 'net_id', 'encryption_type', 'encryption_key', 'feature_bits',
                                     'default_group', 'default_ttl', 'default_delay',
                                     'default_cb_retries', 'default_cb_timeout'])


class Network(object):
    """Parameters defining a SNAP Network

    Networks define 'How do I talk to a SNAP network?'
    """

    def __init__(self, name='default'):
        self.name = name
        self.channel = 4
        self.net_id = '0x1c2c'
        self.encryption_type = 'None'
        self.encryption_key = ''
        self.second_crc = False
        self.packet_crc = False
        self.default_group = 1
        self.default_ttl = 5
        self.default_delay = 50
        self.default_cb_retries = 3
        self.default_cb_timeout = 2

    def __repr__(self):
        return """
{type:s}({name:s}: c={channel:d},n={network:s},e={encryption:s},s={second!s},p={packet!s},
mcast=({group:d}, {ttl:d}, {delay:d}),callback=({retries:d}, {timeout:d})""".format(
            type=self.__class__.__name__,
            name=self.name,
            channel=self.channel,
            network=self.net_id,
            encryption=self.encryption_type,
            second=self.second_crc,
            packet=self.packet_crc,
            group=self.default_group,
            ttl=self.default_ttl,
            delay=self.default_delay,
            retries=self.default_cb_retries,
            timeout=self.default_cb_timeout).replace('\n', '')

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def default():
        return Network()

    @staticmethod
    def list():
        user_cfg = ConfigObj(user_network_path())
        return user_cfg.sections

    @staticmethod
    def delete(name):
        user_cfg = ConfigObj(user_network_path())
        user_cfg.pop(name)
        user_cfg.write()

    @classmethod
    def load(cls, name):
        network = cls(name)
        log.debug("Reading Network: %s", name)

        cfg = ConfigObj(user_network_path())[name]

        # Handle translation from old feature_bit format to crc format
        if cfg.get('feature_bits'):
            cfg['second_crc'] = bool(int(cfg['feature_bits'], 16) & 0x0100)
            cfg['packet_crc'] = bool(int(cfg['feature_bits'], 16) & 0x0400)
            cfg.pop('feature_bits')

        vars(network).update(cfg)
        # Decode the encryption key
        network.encryption_key = network.encryption_key.decode('string_escape')
        network.validate()

        return network

    def save(self):
        """Save the network"""
        self.validate()

        user_cfg = ConfigObj(user_network_path())

        user_cfg[self.name] = vars(self)
        # Don't write the name as a value in the section
        del user_cfg[self.name]['name']

        # Encode the encryption key
        user_cfg[self.name]['encryption_key'] = self.encryption_key.encode("string_escape")

        user_cfg.write()

    def as_sc_params(self):
        """Converts Network to SNAP Connect-ready parameters"""

        crypto_type = 0
        if self.encryption_type.lower() in ("aes-128", '1'):
            crypto_type = 1
        elif self.encryption_type.lower() in ("basic", '2'):
            crypto_type = 2

        second = 0x0100 if self.second_crc else 0
        packet = 0x0400 if self.packet_crc else 0
        feature_bits = second | packet

        return SCNetwork(self.channel, int(self.net_id, 16),
                         crypto_type, self.encryption_key, feature_bits,
                         self.default_group, self.default_ttl, self.default_delay,
                         self.default_cb_retries, self.default_cb_timeout)

    @staticmethod
    def as_bool(var):
        if isinstance(var, bool):
            return var
        if isinstance(var, int):
            return bool(var)
        if isinstance(var, str):
            if var.lower() == 'true':
                return True
            if var.lower() == 'false':
                return False
        raise ValueError('{var} was not able to be type-cast as a bool.'.format(var=var))

    def validate(self):
        """Validate the network against schema."""
        schema = Schema({'name': And(Use(str.strip), And(Regex(VALID_NAME_RE), lambda x: x is not None),
                                     error="Networks must have a valid name"),
                         Optional('channel', default=4): And(Use(int), lambda n: 0 <= n <= 64),
                         Optional('net_id', default='0x1c2c'):
                             Regex(VALID_NET_ID_RE, error="Invalid Network Parameter net_id = %r: Network ID must be 0x####" % self.net_id),
                         Optional('encryption_type', default='None'):
                             lambda x: x in ['None', 'Basic', 'AES-128', '0', '1', '2'],
                         Optional('encryption_key', default=''):
                             And(Use(str), lambda x: (len(x) == 0) or (len(x) == 16)),
                         Optional('second_crc', default=False): Use(Network.as_bool),
                         Optional('packet_crc', default=False): Use(Network.as_bool),
                         Optional('default_group', default=1): And(Use(int), lambda x: 0 < x <= 65535),
                         Optional('default_ttl', default=5): And(Use(int), lambda x: 0 < x <= 255),
                         Optional('default_delay', default=50): And(Use(int), lambda x: 0 <= x <= 255),
                         Optional('default_cb_retries', default=3): And(Use(int), lambda x: 0 <= x <= 255),
                         Optional('default_cb_timeout', default=2): And(Use(int), lambda x: 0 < x <= 255)
                         })

        vars(self).update(schema.validate(vars(self)))
