from collections import namedtuple
import binascii
import logging
import os
import re

from configobj import ConfigObj
from schema import Schema, Use, And, Or, Optional, Regex

log = logging.getLogger(__name__)

VALID_NAME_RE = re.compile(r"^\S+$")
VALID_COM_DEVICE_RE = re.compile(r"^COM([0-9]+)$", re.IGNORECASE)  # Serial on Windows
VALID_USB_DEVICE_RE = re.compile(r"^USB([0-9]+)$", re.IGNORECASE)  # SS100 - USB
VALID_SS_DEVICE_RE = re.compile(r"^SnapStick([0-9]+)$", re.IGNORECASE)  # SS200 - USB
VALID_RS232_DEVICE_RE = re.compile(r"^/dev/(snap|tty\.usbserial|ttyUSB|ttyAMA|ttyO|ttyS|tty\.SLAB_USBtoUART).*$")  # Serial on Posix
VALID_MESHBOARD2_RE = re.compile(r"^MeshBoard2:([0-9]+\.[0-9]+)$", re.IGNORECASE)  # MeshBoard2 - USB
VALID_ADDRESS_RE = re.compile(r"^[0-9A-Fa-f]{6}$")

SERIAL_TYPE_RS232 = 1
SERIAL_TYPE_SNAPSTICK100 = 2
SERIAL_TYPE_SNAPSTICK200 = 5
SERIAL_TYPE_MESH_BOARD2 = 11


def _env_override():
    """Try to find SNAP_PROFILE_PATH env var"""

    try:
        return os.environ["SNAP_PROFILE_PATH"]
    except KeyError:
        return None


def user_profile_path():
    """Check for the environment var override or use the default"""
    return _env_override() or os.path.expanduser(os.path.join("~", ".snap", "profiles.cfg"))


class Profile(object):
    """Parameters defining a SNAP Profile.

    Profiles define 'How do I connect to a SNAP network?'
    """

    def __init__(self, name="default"):
        self.name = name

    def __repr__(self):
        return "%s(%s)" % (self.__class__.__name__, self.name)

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.__dict__ == other.__dict__

    def __ne__(self, other):
        return not self.__eq__(other)

    @staticmethod
    def list():
        user_cfg = ConfigObj(user_profile_path())
        return user_cfg.sections

    @staticmethod
    def delete(name):
        user_cfg = ConfigObj(user_profile_path())
        user_cfg.pop(name)
        user_cfg.write()

    @classmethod
    def default(cls):
        return cls()

    @staticmethod
    def load(name):
        log.debug("Reading Profile: %s", name)

        # So...this is awkward. Need to read enough of the profile
        # to find out what kind it is, then switch on that

        raw_cfg = ConfigObj(user_profile_path())[name]

        log.debug("Raw Config: %r", raw_cfg)

        # sigh...
        if raw_cfg["kind"] == "serial":
            cls = SerialProfile
        elif raw_cfg["kind"] == "tcp":
            cls = TcpProfile
        elif raw_cfg["kind"] == "mqtt":
            cls = MqttProfile
        else:
            raise ValueError("Profile kind must be 'serial', 'tcp', or 'mqtt'")

        log.debug("Profile Class: %r", cls)

        profile = cls(name)
        vars(profile).update(**raw_cfg)
        profile.validate()

        return profile

    def save(self):
        self.validate()

        user_cfg = ConfigObj(user_profile_path())

        user_cfg[self.name] = vars(self)
        # Don't write the name as a value in the section
        del user_cfg[self.name]["name"]

        user_cfg.write()

    def validate(self):
        """Validate the Profile with it's associated schema validation"""
        raise NotImplementedError


SCSerialParams = namedtuple("SCSerialParams", ["type", "port", "address", "speed"])
SCTcpParams = namedtuple("SCTcpParams", ["host", "port", "ssl", "username", "password", "address"])


class SerialProfile(Profile):
    """Serial Profile"""

    def __init__(self, name="default"):
        super(SerialProfile, self).__init__(name)

        self.kind = "serial"
        self.device = "/dev/snap1"
        self.address = ""
        self.network = "default"
        self.speed = 38400

    def as_sc_params(self):
        """Translate from Profile to SC Params"""
        serial_type, serial_port = self._match_device()
        return SCSerialParams(serial_type, serial_port, binascii.unhexlify(self.address) if self.address else None, self.speed)

    def _match_device(self):
        """Figure out what the SNAPconnect serial type and port are from our device string"""
        # Turn our device string into what SC expects
        usb_match = VALID_USB_DEVICE_RE.match(self.device)
        if usb_match:
            # USB SNAPStick 100, aka old paddleboard.
            # Why aren't you in VCP mode???
            serial_type = SERIAL_TYPE_SNAPSTICK100
            # USB0 -> port 0
            serial_port = int(usb_match.groups()[0])

            return serial_type, serial_port

        ss_match = VALID_SS_DEVICE_RE.match(self.device)
        if ss_match:
            serial_type = SERIAL_TYPE_SNAPSTICK200
            # SnapStick0 -> port 0
            serial_port = int(ss_match.groups()[0])

            return serial_type, serial_port

        com_match = VALID_COM_DEVICE_RE.match(self.device)
        if com_match:
            serial_type = SERIAL_TYPE_RS232
            # COM1 -> port 0
            serial_port = int(com_match.groups()[0]) - 1

            return serial_type, serial_port

        dev_match = VALID_RS232_DEVICE_RE.match(self.device)
        if dev_match:
            serial_type = SERIAL_TYPE_RS232
            serial_port = self.device

            return serial_type, serial_port

        meshboard2_match = VALID_MESHBOARD2_RE.match(self.device)
        if meshboard2_match:
            serial_type = SERIAL_TYPE_MESH_BOARD2
            serial_port = meshboard2_match.groups()[0]

            return serial_type, serial_port

        raise ValueError("Invalid Serial Profile Device: %r" % self.device)

    @staticmethod
    def _validate_device(device):
        if not (VALID_COM_DEVICE_RE.match(device) or
                VALID_USB_DEVICE_RE.match(device) or
                VALID_SS_DEVICE_RE.match(device) or
                VALID_RS232_DEVICE_RE.match(device) or
                VALID_MESHBOARD2_RE.match(device)):
            raise ValueError()
        return device

    def validate(self):
        """Validate v0 profile using schema"""
        schema = Schema({"name": And(Use(str.strip), And(Regex(VALID_NAME_RE), lambda x: x is not None),
                                     error="Networks must have a valid name"),
                         Optional("kind", default="serial"): And(Use(str), lambda x: x in ["serial"]),
                         Optional("device", default="/dev/snap1"):
                             Use(SerialProfile._validate_device,
                                 error="Invalid Serial Profile Device = %r" % self.device),
                         Optional("address", default=""):
                             Or(Regex(VALID_ADDRESS_RE), lambda x: x == "",
                                error="Invalid Profile Address = %r" % self.address),
                         Optional("network", default="default"):
                             Or(Regex(VALID_NAME_RE), lambda x: x == "",
                                error="Invalid Profile Network = %r: Must be a valid identifier" % self.network),
                         Optional("speed", default=38400):
                             And(Use(int), lambda x: 2 <= x <= 0xFFFF or x == 115200,
                                 error="Invalid Speed = %r: Must be 2-65535 or 115200" % self.speed),
                         })

        vars(self).update(schema.validate(vars(self)))


class TcpProfile(Profile):
    """TCP Profile"""

    def __init__(self, name="default"):
        super(TcpProfile, self).__init__(name)

        self.kind = "tcp"
        self.host = "127.0.0.1"
        self.port = 48625
        self.ssl = False
        self.username = "public"
        self.password = "public"
        self.address = ""
        self.network = "default"

    def as_sc_params(self):
        return SCTcpParams(self.host, self.port, self.ssl, self.username, self.password,
                           binascii.unhexlify(self.address) if self.address else None)

    @staticmethod
    def as_bool(var):
        if isinstance(var, bool):
            return var
        if isinstance(var, int):
            return bool(var)
        if isinstance(var, str):
            if var.lower() == "true":
                return True
            if var.lower() == "false":
                return False
        raise ValueError("{var} was not able to be type-cast as a bool.".format(var=var))

    def validate(self):
        """Validate v0 profile using schema"""
        schema = Schema({"name": And(Use(str.strip), And(Regex(VALID_NAME_RE), lambda x: x is not None),
                                     error="Networks must have a valid name"),
                         Optional("kind", default="tcp"): And(Use(str), lambda x: x in ["tcp"]),
                         Optional("host", default="127.0.0.1"): And(Use(str), lambda x: len(x) >= 1),
                         Optional("port", default=48625): And(Use(int), lambda x: 0 <= x <= 65535),
                         Optional("ssl", default=False): Use(TcpProfile.as_bool),
                         Optional("username", default="public"): Use(str),
                         Optional("password", default="public"): Use(str),
                         Optional("address", default=""):
                             Or(Regex(VALID_ADDRESS_RE), lambda x: x == "",
                                error="Invalid Profile Address = %r" % self.address),
                         Optional("network", default="default"):
                             Or(Regex(VALID_NAME_RE), lambda x: x == "",
                                error="Invalid Profile Network = %r: Must be a valid identifier" % self.network),
                         })

        vars(self).update(schema.validate(vars(self)))


class MqttProfile(Profile):
    """MQTT Profile"""

    def __init__(self, name="default"):
        super(MqttProfile, self).__init__(name)
        self.kind = "mqtt"
        self.host = "127.0.0.1"
        self.port = 1883
        self.username = ""
        self.password = ""
        self.address = "000080"
        self.ca_certs = ""
        self.network = "default"

    def validate(self):
        """Validate v0 profile using schema"""
        schema = Schema({"name": And(Use(str.strip), And(Regex(VALID_NAME_RE), lambda x: x is not None),
                                     error="Networks must have a valid name"),
                         Optional("kind", default="mqtt"): And(str, lambda x: x in ["mqtt"]),
                         Optional("host", default="127.0.0.1"): And(str, lambda x: len(x) >= 1),
                         Optional("port", default=1883): And(Use(int), lambda x: 0 <= x <= 65535),
                         Optional("address", default=""):
                             Regex(VALID_ADDRESS_RE, error="Invalid Profile Address = %r" % self.address),
                         Optional("ca_certs", default=""): Use(str),
                         Optional("username", default=""): Use(str),
                         Optional("password", default=""): Use(str),
                         Optional("network", default="default"):
                             Or(Regex(VALID_NAME_RE), lambda x: x == "",
                                error="Invalid Profile Network = %r: Must be a valid identifier" % self.network),
                         })

        vars(self).update(schema.validate(vars(self)))
